# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

import xml.parsers.expat
import xmlrpc.client
from typing import Any
from xmlrpc.client import Fault, ResponseError

import defusedxml.xmlrpc

from modernrpc.exceptions import RPCInsecureRequest, RPCInvalidRequest, RPCMarshallingError, RPCParseError
from modernrpc.types import RpcErrorResult
from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult


class PythonXmlRpcBackend:
    """xml-rpc serializer and deserializer based on python builtin xmlrpc module"""

    def __init__(self, load_kwargs: dict[str, Any] | None = None, dump_kwargs: dict[str, Any] | None = None):
        defusedxml.xmlrpc.monkey_patch()

        self.load_kwargs = load_kwargs or {}
        self.load_kwargs.setdefault("use_datetime", True)
        self.load_kwargs.setdefault("use_builtin_types", True)

        self.dump_kwargs = dump_kwargs or {}
        self.dump_kwargs.setdefault("allow_none", True)

    def loads(self, data: str) -> XmlRpcRequest:
        try:
            params, method_name = xmlrpc.client.loads(data, **self.load_kwargs)
        except xml.parsers.expat.ExpatError as exc:
            raise RPCParseError(str(exc)) from exc
        except (ResponseError, TypeError) as exc:
            raise RPCInvalidRequest(str(exc)) from exc
        except defusedxml.DefusedXmlException as exc:
            raise RPCInsecureRequest(str(exc)) from exc

        if not method_name:
            raise RPCInvalidRequest("Unable to find method name", data=data)

        return XmlRpcRequest(method_name.strip(), list(params))

    def dumps(self, result: XmlRpcResult) -> str:
        result_data = Fault(result.code, result.message) if isinstance(result, RpcErrorResult) else (result.data,)
        try:
            return xmlrpc.client.dumps(result_data, methodresponse=True, **self.dump_kwargs)
        except Exception as exc:
            raise RPCMarshallingError(result.data, exc) from exc



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any

import lxml.etree
from lxml.etree import Element, SubElement, _Element

from modernrpc.exceptions import RPCInvalidRequest, RPCMarshallingError, RPCParseError
from modernrpc.xmlrpc.backends.marshalling import EtreeElementMarshaller, EtreeElementUnmarshaller

if TYPE_CHECKING:
    from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult


class LxmlBackend:
    """xml-rpc serializer and deserializer based on the third-party lxml library"""

    def __init__(self, load_kwargs: dict[str, Any] | None = None, dump_kwargs: dict[str, Any] | None = None):
        self.load_kwargs = load_kwargs or {}
        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def unmarshaller(self):
        return EtreeElementUnmarshaller[_Element]()

    @cached_property
    def marshaller(self):
        return EtreeElementMarshaller[_Element](Element, SubElement)

    def loads(self, data: str) -> XmlRpcRequest:
        # Create a secure parser that disables DTDs, external entities, and entity expansion
        parser = lxml.etree.XMLParser(
            resolve_entities=False,  # Prevent entity expansion
            no_network=True,  # Prevent network access
            dtd_validation=False,  # Disable DTD validation
            load_dtd=False,  # Disable DTD loading
            huge_tree=False,  # Prevent the billion-laugh attack
        )

        try:
            root_obj: _Element = lxml.etree.fromstring(data, parser)
        except lxml.etree.XMLSyntaxError as exc:
            raise RPCParseError(str(exc)) from exc

        try:
            return self.unmarshaller.element_to_request(root_obj)
        except Exception as exc:
            raise RPCInvalidRequest(str(exc)) from exc

    def dumps(self, result: XmlRpcResult) -> str:
        """Serialize an XmlRpcResult to an XML string."""
        try:
            root = self.marshaller.result_to_element(result)
        except Exception as exc:
            raise RPCMarshallingError(result.data, exc) from exc

        return lxml.etree.tostring(root, encoding="utf-8", xml_declaration=True).decode("utf-8")



from __future__ import annotations

import asyncio

from modernrpc import Protocol, RpcNamespace, RpcRequestContext
from modernrpc.config import settings
from modernrpc.exceptions import RPCInvalidParams
from modernrpc.types import RpcErrorResult
from modernrpc.xmlrpc.handler import XmlRpcRequest

system = RpcNamespace()


@system.register_procedure(name="listMethods", context_target="_ctx")
def __system_list_methods(_ctx: RpcRequestContext):
    """Returns a list of all methods available in the current entry point"""
    server = _ctx.server
    return list(server.procedures.keys())


@system.register_procedure(name="methodSignature", context_target="_ctx")
def __system_method_signature(method_name: str, _ctx: RpcRequestContext):
    """
    Returns an array describing the signature of the given method name.

    The result is an array with:
     - Return type as first elements
     - Types of method arguments from element 1 to N
    :param method_name: Name of a method available for current entry point (and protocol)
    :param _ctx: Request context for this call
    :return: An array of arrays describing types of return values and method arguments
    """
    server = _ctx.server
    wrapper = server.get_procedure_wrapper(method_name, _ctx.protocol)
    # See http://xmlrpc-c.sourceforge.net/introspection.html
    undefined = "undef"
    return_type = wrapper.return_doc.get("type") or undefined
    args_types = [arg_doc.get("type") or undefined for arg_doc in wrapper.args_doc.values()]

    return [[return_type, *args_types]]


@system.register_procedure(name="methodHelp", context_target="_ctx")
def __system_method_help(method_name: str, _ctx: RpcRequestContext):
    """
    Returns the documentation of the given method name.

    :param method_name: Name of a method available for the current entry point (and protocol)
    :param _ctx: Request context for this call
    :return: Documentation text for the RPC method
    """
    server = _ctx.server
    method = server.get_procedure_wrapper(method_name, _ctx.protocol)
    return method.raw_docstring


if settings.MODERNRPC_XMLRPC_ASYNC_MULTICALL:

    @system.register_procedure(name="multicall", protocol=Protocol.XML_RPC, context_target="_ctx")
    async def __system_multi_call(calls: list, _ctx: RpcRequestContext):
        """
        Call multiple RPC methods at once, using asyncio.gather().

        :param calls: An array of struct like {"methodName": string, "params": array }
        :param _ctx: Request context for this call
        :return: An array containing the result of each procedure call
        """
        if not isinstance(calls, list):
            raise RPCInvalidParams(f"system.multicall first argument should be a list, {type(calls).__name__} given.")

        requests = (XmlRpcRequest(call.get("methodName"), call.get("params")) for call in calls)
        results = await asyncio.gather(*(_ctx.handler.aprocess_single_request(request, _ctx) for request in requests))

        return [
            {"faultCode": result.code, "faultString": result.message}
            if isinstance(result, RpcErrorResult)
            else (result.data,)
            for result in results
        ]
else:

    @system.register_procedure(name="multicall", protocol=Protocol.XML_RPC, context_target="_ctx")
    def __system_multi_call(calls: list, _ctx: RpcRequestContext):
        """
        Call multiple RPC methods at once.

        :param calls: An array of struct like {"methodName": string, "params": array }
        :param _ctx: Request context for this call
        :return: An array containing the result of each procedure call
        """
        if not isinstance(calls, list):
            raise RPCInvalidParams(f"system.multicall first argument should be a list, {type(calls).__name__} given.")

        requests = (XmlRpcRequest(call.get("methodName"), call.get("params")) for call in calls)
        results = (_ctx.handler.process_single_request(request, _ctx) for request in requests)

        return [
            {"faultCode": result.code, "faultString": result.message}
            if isinstance(result, RpcErrorResult)
            else (result.data,)
            for result in results
        ]



from __future__ import annotations

import logging
from collections import OrderedDict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Iterable

from asgiref.sync import async_to_sync, iscoroutinefunction, sync_to_async
from django.utils.functional import cached_property

from modernrpc import Protocol
from modernrpc.constants import NOT_SET
from modernrpc.exceptions import (
    AuthenticationError,
    RPCInvalidParams,
)
from modernrpc.helpers import check_flags_compatibility, ensure_sequence
from modernrpc.introspection import DocstringParser, Introspector

if TYPE_CHECKING:
    from django.http import HttpRequest

    from modernrpc.handler import RpcHandler
    from modernrpc.server import RpcServer
    from modernrpc.types import AuthPredicateType


logger = logging.getLogger(__name__)


@dataclass
class RpcRequestContext:
    """Wraps all information needed to call a procedure. Instances of this class are created before call
    and may be used to populate kwargs dict in rpc method."""

    request: HttpRequest
    server: RpcServer
    handler: RpcHandler
    protocol: Protocol
    auth_result: Any = None


class ProcedureWrapper:
    """
    Wraps a python global function to be used to extract information and call the concrete procedure.
    """

    def __init__(
        self,
        func_or_coro: Callable,
        name: str | None = None,
        protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        context_target: str | None = None,
    ) -> None:
        # Store the reference to the registered function
        self.func_or_coro = func_or_coro

        # @decorator parameters
        self.name = name or func_or_coro.__name__
        self.protocol = protocol
        self.context_target = context_target

        self.auth = auth

    @cached_property
    def doc_parser(self) -> DocstringParser:
        return DocstringParser(self.func_or_coro)

    @cached_property
    def introspector(self) -> Introspector:
        return Introspector(self.func_or_coro)

    def __repr__(self) -> str:
        return f'ModernRPC Procedure "{self.name}"'

    def __str__(self) -> str:
        return f"{self.name}({', '.join(self.args)})"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, ProcedureWrapper):
            return NotImplemented
        return (
            self.func_or_coro == other.func_or_coro
            and self.name == other.name
            and self.protocol == other.protocol
            and self.auth == other.auth
        )

    def __hash__(self):
        return hash(self.func_or_coro)

    def check_permissions(self, request: HttpRequest) -> Any:
        """
        Call each predicate associated with the procedure to check authentication.
        If any of the predicate returns a truthy value, it is returned as result.
        In other cases, raise an AuthenticationFailed exception.
        """
        if self.auth == NOT_SET or not self.auth:
            return True

        for callback in ensure_sequence(self.auth):
            if result := callback(request):
                return result

        raise AuthenticationError(self.name)

    def execute(
        self,
        context: RpcRequestContext,
        args: Iterable[Any] | None = None,
        kwargs: dict | None = None,
    ) -> Any:
        args = args or []
        kwargs = kwargs or {}

        try:
            auth_result = self.check_permissions(context.request)
        except Exception as exc:
            raise AuthenticationError(self.name) from exc

        # If the remote procedure requested access to context data, provide it into proper kwargs key
        if self.context_target:
            context.auth_result = auth_result
            kwargs[self.context_target] = context

        logger.debug("Params: args = %s - kwargs = %s", args, kwargs)

        function = async_to_sync(self.func_or_coro) if iscoroutinefunction(self.func_or_coro) else self.func_or_coro

        try:
            return function(*args, **kwargs)

        except TypeError as exc:
            # If given params cannot be transmitted properly to the procedure function
            raise RPCInvalidParams(str(exc)) from None

    async def aexecute(
        self,
        context: RpcRequestContext,
        args: Iterable[Any] | None = None,
        kwargs: dict | None = None,
    ) -> Any:
        args = args or []
        kwargs = kwargs or {}

        try:
            auth_result = self.check_permissions(context.request)
        except Exception as exc:
            raise AuthenticationError(self.name) from exc

        # If the remote procedure requested access to context data, provide it into proper kwargs key
        if self.context_target:
            context.auth_result = auth_result
            kwargs[self.context_target] = context

        logger.debug("Params: args = %s - kwargs = %s", args, kwargs)

        coro = self.func_or_coro if iscoroutinefunction(self.func_or_coro) else sync_to_async(self.func_or_coro)

        try:
            return await coro(*args, **kwargs)
        except TypeError as exc:
            # If given params cannot be transmitted properly to the procedure function
            raise RPCInvalidParams(str(exc)) from None

    @cached_property
    def args(self) -> list[str]:
        """Method arguments"""
        return [arg for arg in self.introspector.args if arg != self.context_target]

    @cached_property
    def raw_docstring(self) -> str:
        """Method docstring, as raw text"""
        return self.doc_parser.text_documentation

    @cached_property
    def html_doc(self) -> str:
        """Methods docstring, as HTML"""
        return self.doc_parser.html_documentation

    @cached_property
    def args_doc(self) -> OrderedDict[str, dict[str, str]]:
        """Build an OrderedDict mapping each method argument with its
        corresponding type (from typehint or doctype) and documentation."""
        return OrderedDict(
            {
                arg: {
                    "type": self.introspector.args_types.get(arg, "") or self.doc_parser.args_types.get(arg, ""),
                    "text": self.doc_parser.args_doc.get(arg, ""),
                }
                for arg in self.introspector.args
                if arg != self.context_target
            }
        )

    @cached_property
    def return_doc(self) -> dict[str, str]:
        """Build a dict for the method's return type and documentation"""
        return {
            "type": self.introspector.return_type or self.doc_parser.return_type,
            "text": self.doc_parser.return_doc,
        }

    @cached_property
    def is_available_in_xml_rpc(self):
        return check_flags_compatibility(self.protocol, Protocol.XML_RPC)

    @cached_property
    def is_available_in_json_rpc(self):
        return check_flags_compatibility(self.protocol, Protocol.JSON_RPC)



import logging

from django.apps import AppConfig

logger = logging.getLogger(__name__)


class ModernRpcConfig(AppConfig):
    name = "modernrpc"
    verbose_name = "Django Modern RPC"

    def ready(self): ...



MODERNRPC_DEFAULT_ENCODING = "utf-8"

# List of handler classes used by default in any ``RPCEntryPoint`` instance
MODERNRPC_HANDLERS = [
    "modernrpc.jsonrpc.handler.JsonRpcHandler",
    "modernrpc.xmlrpc.handler.XmlRpcHandler",
]

MODERNRPC_JSON_DESERIALIZER = {"class": "modernrpc.jsonrpc.backends.json.PythonJsonBackend", "kwargs": {}}
MODERNRPC_JSON_SERIALIZER = {"class": "modernrpc.jsonrpc.backends.json.PythonJsonBackend", "kwargs": {}}
MODERNRPC_XML_DESERIALIZER = {"class": "modernrpc.xmlrpc.backends.xmlrpc.PythonXmlRpcBackend", "kwargs": {}}
MODERNRPC_XML_SERIALIZER = {"class": "modernrpc.xmlrpc.backends.xmlrpc.PythonXmlRpcBackend", "kwargs": {}}

# Configure the format of the docstring used to document your RPC methods.
# Possible values are: '', 'rst' or 'md'
MODERNRPC_DOC_FORMAT = ""

MODERNRPC_XMLRPC_ASYNC_MULTICALL = False



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic

from modernrpc.exceptions import RPCMethodNotFound
from modernrpc.types import RequestType, RpcErrorResult, RpcSuccessResult

if TYPE_CHECKING:
    from http import HTTPStatus
    from typing import Any, ClassVar

    from django.http import HttpRequest

    from modernrpc import Protocol, RpcRequestContext


logger = logging.getLogger(__name__)


class RpcHandler(ABC, Generic[RequestType]):
    """Base class for concrete RPC Handlers. Provide an interface as well as some common methods implementations."""

    protocol: Protocol
    valid_content_types: ClassVar[list[str]]
    response_content_type: str
    success_result_type: type[RpcSuccessResult[RequestType]]
    error_result_type: type[RpcErrorResult[RequestType]]

    @classmethod
    def can_handle(cls, request: HttpRequest) -> bool:
        """
        Return True if this instance can handle the given request.

        The default implementation will check Content-Type for a supported value
        """
        return getattr(request, "content_type", "").lower() in cls.valid_content_types

    # @abc.abstractmethod
    def build_success_result(self, request: RequestType, data: Any) -> RpcSuccessResult[RequestType]:
        return self.success_result_type(request=request, data=data)

    # @abc.abstractmethod
    def build_error_result(
        self, request: RequestType, code: int, message: str, data: Any = None
    ) -> RpcErrorResult[RequestType]:
        return self.error_result_type(request=request, code=code, message=message, data=data)

    def process_single_request(
        self, rpc_request: RequestType, context: RpcRequestContext
    ) -> RpcSuccessResult[RequestType] | RpcErrorResult[RequestType]:
        """Check and call the RPC method, based on the given request dict."""

        try:
            wrapper = context.server.get_procedure_wrapper(rpc_request.method_name, self.protocol)

        except RPCMethodNotFound as exc:
            # exc variable cannot be reused here without triggering an error with static type checker
            # In this particular case, an error handler may decide to convert the given RPCMethodNotFound to a more
            # generic RPCException (or one of its subclasses).
            excp = context.server.on_error(exc, context)
            return self.build_error_result(request=rpc_request, code=excp.code, message=excp.message, data=excp.data)

        try:
            result_data = wrapper.execute(context, rpc_request.args, getattr(rpc_request, "kwargs", None))

        except Exception as exc:
            exc = context.server.on_error(exc, context)
            return self.build_error_result(request=rpc_request, code=exc.code, message=exc.message, data=exc.data)

        return self.build_success_result(rpc_request, result_data)

    async def aprocess_single_request(
        self, rpc_request: RequestType, context: RpcRequestContext
    ) -> RpcSuccessResult[RequestType] | RpcErrorResult[RequestType]:
        """
        Asynchronous version of process_single_request().
        Check and call the RPC method, based on the given request dict.
        """

        try:
            wrapper = context.server.get_procedure_wrapper(rpc_request.method_name, self.protocol)

        except RPCMethodNotFound as exc:
            # exc variable cannot be reused here without triggering an error with static type checker
            # In this particular case, an error handler may decide to convert the given RPCMethodNotFound to a more
            # generic RPCException (or one of its subclasses).
            excp = context.server.on_error(exc, context)
            return self.build_error_result(request=rpc_request, code=excp.code, message=excp.message, data=excp.data)

        try:
            result_data = await wrapper.aexecute(context, rpc_request.args, getattr(rpc_request, "kwargs", None))

        except Exception as exc:
            exc = context.server.on_error(exc, context)
            return self.build_error_result(request=rpc_request, code=exc.code, message=exc.message, data=exc.data)

        return self.build_success_result(rpc_request, result_data)

    @abstractmethod
    def process_request(self, request_body: str, context: RpcRequestContext) -> str | tuple[HTTPStatus, str]:
        """
        Fully process a request. Return the str content ready to be sent as HttpResponse.

        This is the only method the view is supposed to call after choosing the right handler.
        Concrete implementation must perform all work here, according to its protocol specifications.

        Important: This method should delegate concrete procedure execution to process_single_request(). This will
        help to implement multi-request (JSON-RPC batch / XML-RPC system.multicall) features.

        Implementations of this method must ensure no exception is raised from here. All code must be secured
        to return a proper RPC error response on any exception.
        """

    @abstractmethod
    async def aprocess_request(self, request_body: str, context: RpcRequestContext) -> str | tuple[HTTPStatus, str]:
        """
        Asynchronous version of process_request(). It takes the same arguments and returns the same result.
        Delegates its work to aprocess_single_request() instead of process_single_request() for async support.
        """



from __future__ import annotations

from typing import TYPE_CHECKING, Iterable, Protocol

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcRequest, JsonRpcResult


class JsonRpcDeserializer(Protocol):
    def __init__(self, **kwargs): ...

    def loads(self, data: str) -> JsonRpcRequest | list[JsonRpcRequest]: ...


class JsonRpcSerializer(Protocol):
    def __init__(self, **kwargs): ...

    def dumps(self, result: JsonRpcResult | Iterable[JsonRpcResult]) -> str: ...



import functools

import django
from django.http import HttpResponseNotAllowed
from django.utils.log import log_response
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

if django.VERSION >= (5, 0):
    # Django updated its decorators to support wrapping asynchronous method in release 5.0
    # REF: https://docs.djangoproject.com/en/5.2/releases/5.0/#decorators

    async_csrf_exempt = csrf_exempt
    async_require_post = require_POST

else:
    # Fore Django version < 5.0, we partially redefine decorators to support async view

    def async_csrf_exempt(view):
        view.csrf_exempt = True
        return view

    def async_require_post(func):
        @functools.wraps(func)
        async def inner(request, *args, **kwargs):
            if request.method != "POST":
                response = HttpResponseNotAllowed([""])
                log_response(
                    "Method Not Allowed (%s): %s",
                    request.method,
                    request.path,
                    response=response,
                    request=request,
                )
                return response
            return await func(request, *args, **kwargs)

        return inner



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterable

import rapidjson
from rapidjson import JSONDecodeError

from modernrpc.exceptions import RPCMarshallingError, RPCParseError
from modernrpc.jsonrpc.backends.marshalling import Marshaller, Unmarshaller

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcRequest, JsonRpcResult


class RapidJsonBackend:
    """json-rpc serializer and deserializer based on the third-party simplejson library"""

    def __init__(self, load_kwargs: dict[str, Any] | None = None, dump_kwargs: dict[str, Any] | None = None):
        self.load_kwargs = load_kwargs or {}
        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def unmarshaller(self):
        return Unmarshaller()

    @cached_property
    def marshaller(self):
        return Marshaller()

    def loads(self, data: str) -> JsonRpcRequest | list[JsonRpcRequest]:
        try:
            structured_data: list[dict] | dict[str, Any] = rapidjson.loads(data, **self.load_kwargs)
        except JSONDecodeError as exc:
            raise RPCParseError(str(exc), data=exc) from exc

        return self.unmarshaller.dict_to_request(structured_data)

    def dumps(self, result: JsonRpcResult | Iterable[JsonRpcResult]) -> str:
        structured_data = self.marshaller.result_to_dict(result)
        try:
            return rapidjson.dumps(structured_data, **self.dump_kwargs)
        except (TypeError, UnicodeDecodeError) as exc:
            raise RPCMarshallingError(structured_data, exc) from exc



# JSON-RPC error codes
# From https://www.jsonrpc.org/specification#error_object
#
# -32700              Parse error 	    Invalid JSON was received by the server.
#                                         An error occurred on the server while parsing the JSON text.
# -32600              Invalid Request 	The JSON sent is not a valid Request object.
# -32601              Method not found 	The method does not exist / is not available.
# -32602              Invalid params 	    Invalid method parameter(s).
# -32603              Internal error 	    Internal JSON-RPC error.
# -32000 to -32099 	Server error 	    Reserved for implementation-defined server-errors.
#
# XML-RPC error codes
# From http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php
#
# -32700 ---> parse error. not well formed
# -32701 ---> parse error. unsupported encoding
# -32702 ---> parse error. invalid character for encoding
# -32600 ---> server error. invalid xml-rpc. not conforming to spec.
# -32601 ---> server error. requested method not found
# -32602 ---> server error. invalid method parameters
# -32603 ---> server error. internal xml-rpc error
# -32500 ---> application error
# -32400 ---> system error
# -32300 ---> transport error
#
# In addition, the range -32099 .. -32000, inclusive is reserved for implementation defined server errors. Server
# errors which do not cleanly map to a specific error defined by this spec should be assigned to a number in this range.
# This leaves the remainder of the space available for application defined errors.

"""
Error handling is fully described in both XML & JSON-RPC standards. Each common error have an associated *faultCode*
and the response format is described, so errors can be handled correctly on the client side.

In django-modern-rpc, all errors are reported using a set of pre-defined exceptions. Thus, in JSON and XML-RPC handlers,
when an exception is caught, the correct error response is returned to the view and transmitted to the client.

This simplifies error management, and allow developers to simply return errors to clients from inside a RPC Method.
The error codes values are defined in:

- http://www.jsonrpc.org/specification#error_object for JSON-RPC
- http://xmlrpc-epi.sourceforge.net/specs/rfc.fault_codes.php for XML-RPC

Pre-defined exceptions uses the following error codes::

    RPC_PARSE_ERROR = -32700
    RPC_INVALID_REQUEST = -32600
    RPC_METHOD_NOT_FOUND = -32601
    RPC_INVALID_PARAMS = -32602
    RPC_INTERNAL_ERROR = -32603

    # Used as minimal value for any custom error returned by the server
    RPC_CUSTOM_ERROR_BASE = -32099
    # Used as maximal value for any custom error returned by the server
    RPC_CUSTOM_ERROR_MAX = -32000

"""

from typing import Any

RPC_PARSE_ERROR = -32700
RPC_INVALID_REQUEST = -32600
RPC_METHOD_NOT_FOUND = -32601
RPC_INVALID_PARAMS = -32602
RPC_INTERNAL_ERROR = -32603

# Used as minimal value for any custom error returned by the server
RPC_CUSTOM_ERROR_BASE = -32099
# Used as maximal value for any custom error returned by the server
RPC_CUSTOM_ERROR_MAX = -32000


class RPCException(Exception):
    """
    This is the base class of all RPC exception. Custom exceptions raised by your RPC methods
    should inherit from RPCException.
    """

    def __init__(self, code: int, message: str, data: Any = None):
        super().__init__(f"[{code}] {message}")
        self.code = code
        self.message = message
        self.data = data


class RPCParseError(RPCException):
    """Raised by handlers if the request can't be read as valid JSON or XML data."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Parse error, unable to read the request: {message}"
        super().__init__(RPC_PARSE_ERROR, err_msg, data)


class RPCInsecureRequest(RPCParseError):
    """Raised in backends if the request is maliciously formed."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Security error: {message}"
        super().__init__(err_msg, data)


class RPCInvalidRequest(RPCException):
    """Raised by handlers if incoming JSON or XML data is not a valid JSON-RPC or XML-RPC data."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Invalid request: {message}"
        super().__init__(RPC_INVALID_REQUEST, err_msg, data)


class RPCMethodNotFound(RPCException):
    """Raised by handlers when no procedure were found with the given name in the server (for the current protocol)"""

    def __init__(self, name: str, data: Any = None):
        err_msg = f'Method not found: "{name}"'
        super().__init__(RPC_METHOD_NOT_FOUND, err_msg, data)


class RPCInvalidParams(RPCException):
    """Raised by handlers if the request params does not match the procedure's expected ones"""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Invalid parameters: {message}"
        super().__init__(RPC_INVALID_PARAMS, err_msg, data)


class RPCInternalError(RPCException):
    """Raised by handlers if any standard exception is raised during the execution of the procedure."""

    def __init__(self, message: str, data: Any = None):
        err_msg = f"Internal error: {message}"
        super().__init__(RPC_INTERNAL_ERROR, err_msg, data)


class RPCMarshallingError(RPCException):
    def __init__(self, data: Any, exc: Exception):
        super().__init__(
            RPC_INTERNAL_ERROR,
            f"Unable to serialize result data: {data}. Original exception: {exc}",
        )


class AuthenticationError(RPCException):
    """Raised when authentication system forbade execution of a RPC Method"""

    def __init__(self, method_name: str):
        super().__init__(
            RPC_INTERNAL_ERROR,
            f'Authentication failed when calling "{method_name}"',
        )


# In 1.0, RPCUnknownMethod were renamed to RPCMethodNotFound
# Set an alias for backward compatibility
RPCUnknownMethod = RPCMethodNotFound



from __future__ import annotations

import base64
from collections import OrderedDict
from datetime import datetime
from typing import Any, Callable, Generic, Iterable, Protocol, TypeVar

from modernrpc.exceptions import RPCInvalidRequest
from modernrpc.helpers import first
from modernrpc.types import RpcErrorResult
from modernrpc.xmlrpc.backends.constants import MAXINT, MININT
from modernrpc.xmlrpc.handler import XmlRpcRequest, XmlRpcResult

# Self is available in typing base module only from Python 3.11
try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

try:
    # types.NoneType is available only with Python 3.10+
    from types import NoneType
except ImportError:
    NoneType = type(None)  # type: ignore[misc]


class ElementTypeProtocol(Protocol, Iterable):
    """
    Base protocol for XML element types. This reflects the API of both the xml.etree and the lxml library.
    Unfortunately, since both libraries share the same interface without inheriting a common base class, we
    have to define our own.
    """

    tag: str
    text: str

    def find(self, path: str, namespaces=None) -> Self | None: ...
    def findall(self, path: str, namespaces=None) -> list[Self]: ...
    def append(self, sub_element: Self) -> None: ...


ElementType = TypeVar("ElementType", bound=ElementTypeProtocol)

LoadFuncType = Callable[[ElementType], Any]
DumpFuncType = Callable[[Any], ElementType]


class EtreeElementUnmarshaller(Generic[ElementType]):
    def __init__(self) -> None:
        self.load_funcs: dict[str, LoadFuncType] = {
            "value": self.load_value,
            "nil": self.load_nil,
            "boolean": self.load_bool,
            "int": self.load_int,
            "i4": self.load_int,
            "double": self.load_float,
            "string": self.load_str,
            "dateTime.iso8601": self.load_datetime,
            "base64": self.load_base64,
            "array": self.load_array,
            "struct": self.load_struct,
        }

    def element_to_request(self, root: ElementType) -> XmlRpcRequest:
        if root.tag != "methodCall":
            raise RPCInvalidRequest("missing methodCall tag", data=root)

        method_name = root.find("./methodName")
        if method_name is None:
            raise RPCInvalidRequest("missing methodCall.methodName tag", data=root)

        params = root.find("./params")
        if params is None:
            return XmlRpcRequest(method_name=self.stripped_text(method_name))

        param_list = params.findall("./param")

        args = [self.dispatch(self.first_child(param)) for param in param_list]
        return XmlRpcRequest(method_name=self.stripped_text(method_name), args=args)

    @staticmethod
    def stripped_text(elt: ElementType) -> str:
        return elt.text.strip() if elt.text else ""

    @staticmethod
    def first_child(elt: ElementType) -> ElementType:
        try:
            return first(elt)
        except IndexError as ie:
            raise RPCInvalidRequest("missing child element", data=elt) from ie

    def dispatch(self, elt: ElementType) -> Any:
        try:
            load_func = self.load_funcs[elt.tag]
        except KeyError as exc:
            raise RPCInvalidRequest(f"Unsupported type {elt.tag}") from exc

        return load_func(elt)

    def load_value(self, element: ElementType) -> Any:
        return self.dispatch(self.first_child(element))

    @staticmethod
    def load_nil(_: ElementType) -> None:
        return None

    def load_int(self, elt: ElementType) -> int:
        return int(self.stripped_text(elt))

    def load_bool(self, elt: ElementType) -> bool:
        value = self.stripped_text(elt)
        if value not in ("0", "1"):
            raise TypeError(f"Invalid boolean value: only 0 and 1 are allowed, found {value}")
        return value == "1"

    def load_float(self, elt: ElementType) -> float:
        return float(self.stripped_text(elt))

    def load_str(self, elt: ElementType) -> str:
        return str(self.stripped_text(elt))

    def load_datetime(self, elt: ElementType) -> datetime:
        return datetime.strptime(self.stripped_text(elt), "%Y%m%dT%H:%M:%S")

    def load_base64(self, elt: ElementType) -> bytes:
        return base64.b64decode(self.stripped_text(elt))

    def load_array(self, elt: ElementType) -> list[Any]:
        return [self.dispatch(value_elt) for value_elt in elt.findall("./data/value")]

    def load_struct(self, elt: ElementType) -> dict[str, Any]:
        member_names_and_values = [self.load_struct_member(member) for member in elt.findall("./member")]
        return dict(member_names_and_values)

    def load_struct_member(self, member_elt: ElementType) -> tuple[str, Any]:
        member_name = member_elt.find("./name")
        if member_name is None:
            raise RPCInvalidRequest("missing member.name tag", data=member_elt)
        value = member_elt.find("./value")
        if value is None:
            raise RPCInvalidRequest("missing member.value tag", data=member_elt)

        return self.stripped_text(member_name), self.dispatch(value)


class EtreeElementMarshaller(Generic[ElementType]):
    def __init__(
        self,
        element_factory: Callable[[str], ElementType],
        sub_element_factory: Callable[[ElementType, str], ElementType],
    ) -> None:
        self.element_factory = element_factory
        self.sub_element_factory = sub_element_factory
        self.dump_funcs: dict[type, DumpFuncType] = {
            NoneType: self.dump_nil,
            bool: self.dump_bool,
            int: self.dump_int,
            float: self.dump_float,
            str: self.dump_str,
            bytes: self.dump_bytearray,
            bytearray: self.dump_bytearray,
            datetime: self.dump_datetime,
            list: self.dump_list,
            tuple: self.dump_list,
            dict: self.dump_dict,
            OrderedDict: self.dump_dict,
        }

    def result_to_element(self, result: XmlRpcResult) -> ElementType:
        """Convert an XmlRpcResult to an XML element."""
        root = self.element_factory("methodResponse")

        if isinstance(result, RpcErrorResult):
            fault = self.sub_element_factory(root, "fault")
            value = self.sub_element_factory(fault, "value")

            struct = self.sub_element_factory(value, "struct")

            # Add faultCode member
            member = self.sub_element_factory(struct, "member")
            name = self.sub_element_factory(member, "name")
            name.text = "faultCode"
            value = self.sub_element_factory(member, "value")
            int_el = self.sub_element_factory(value, "int")
            int_el.text = str(result.code)

            # Add faultString member
            member = self.sub_element_factory(struct, "member")
            name = self.sub_element_factory(member, "name")
            name.text = "faultString"
            value = self.sub_element_factory(member, "value")
            string = self.sub_element_factory(value, "string")
            string.text = result.message
        else:
            params = self.sub_element_factory(root, "params")
            param = self.sub_element_factory(params, "param")
            value = self.sub_element_factory(param, "value")

            # Add the result data
            value.append(self.dispatch(result.data))

        return root

    def dispatch(self, value: Any) -> ElementType:
        """Dispatch a value to the appropriate dump method."""
        try:
            dump_func = self.dump_funcs[type(value)]
        except KeyError as exc:
            raise TypeError(f"Unsupported type: {type(value)}") from exc

        return dump_func(value)

    def dump_nil(self, _: None) -> ElementType:
        return self.element_factory("nil")

    def dump_bool(self, value: bool) -> ElementType:
        boolean = self.element_factory("boolean")
        boolean.text = "1" if value else "0"
        return boolean

    def dump_int(self, value: int) -> ElementType:
        if value > MAXINT or value < MININT:
            raise OverflowError("int value exceeds XML-RPC limits")
        int_el = self.element_factory("int")
        int_el.text = str(value)
        return int_el

    def dump_float(self, value: float) -> ElementType:
        double = self.element_factory("double")
        double.text = str(value)
        return double

    def dump_str(self, value: str) -> ElementType:
        string = self.element_factory("string")
        string.text = value
        return string

    def dump_datetime(self, value: datetime) -> ElementType:
        dt = self.element_factory("dateTime.iso8601")
        dt.text = value.strftime("%04Y%02m%02dT%H:%M:%S")
        return dt

    def dump_bytearray(self, value: bytes | bytearray) -> ElementType:
        b64 = self.element_factory("base64")
        b64.text = base64.b64encode(value).decode()
        return b64

    def dump_dict(self, value: dict) -> ElementType:
        struct = self.element_factory("struct")
        for key, val in value.items():
            member = self.sub_element_factory(struct, "member")
            name = self.sub_element_factory(member, "name")
            name.text = str(key)
            value_el = self.sub_element_factory(member, "value")
            value_el.append(self.dispatch(val))
        return struct

    def dump_list(self, value: list | tuple) -> ElementType:
        array = self.element_factory("array")
        data = self.sub_element_factory(array, "data")
        for val in value:
            value_el = self.sub_element_factory(data, "value")
            value_el.append(self.dispatch(val))
        return array



from __future__ import annotations

import functools
import logging
from typing import TYPE_CHECKING, Callable, Union

from django.utils.module_loading import import_string
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_POST

from modernrpc import Protocol, RpcRequestContext
from modernrpc.compat import async_csrf_exempt, async_require_post
from modernrpc.config import settings
from modernrpc.constants import NOT_SET, SYSTEM_NAMESPACE_DOTTED_PATH
from modernrpc.core import ProcedureWrapper
from modernrpc.exceptions import RPCException, RPCInternalError, RPCMethodNotFound
from modernrpc.helpers import check_flags_compatibility, first_true
from modernrpc.views import handle_rpc_request, handle_rpc_request_async

if TYPE_CHECKING:
    from django.http import HttpRequest

    from modernrpc.handler import RpcHandler
    from modernrpc.types import AuthPredicateType


logger = logging.getLogger(__name__)

RpcErrorHandler = Callable[[BaseException, RpcRequestContext], Union[RPCException, None]]


class RegistryMixin:
    def __init__(self, auth: AuthPredicateType = NOT_SET) -> None:
        self._registry: dict[str, ProcedureWrapper] = {}
        self.auth = auth

    def register_procedure(
        self,
        procedure: Callable | None = None,
        name: str | None = None,
        protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        context_target: str | None = None,
    ) -> Callable:
        """
        Registers a procedure for handling RPC (Remote Procedure Call) requests. This function can be used as a
        decorator to register the given callable function as an RPC procedure.

        :param procedure: A callable function to be registered as an RPC procedure
        :param name: A custom name for the procedure. If not provided, the callable function's name is used by default.
        :param protocol: Specifies the protocol (e.g., JSON-RPC, XML-RPC) under which the procedure can be executed.
                         Defaults: Protocol.ALL
        :param auth: Defines user authentication settings or access rules for the procedure. Defaults to NOT_SET, in
                     which case the server's authentication settings will be used.
        :param context_target: Specify the procedure argument name for accessing the RPC request context.

        :raises ValueError: If a procedure can't be registered
        """

        def decorated(func: Callable) -> Callable:
            if name and name.startswith("rpc."):
                raise ValueError(
                    'According to JSON-RPC specs, method names starting with "rpc." are reserved for system extensions '
                    "and must not be used. See https://www.jsonrpc.org/specification#extensions for more information."
                )

            wrapper = ProcedureWrapper(
                func,
                name,
                protocol=protocol,
                auth=self.auth if auth == NOT_SET else auth,
                context_target=context_target,
            )

            if wrapper.name in self._registry and wrapper != self._registry[wrapper.name]:
                raise ValueError(f"Procedure {wrapper.name} is already registered")

            self._registry[wrapper.name] = wrapper
            logger.debug(f"Registered procedure {wrapper.name}")
            return func

        # If @server.register_procedure() is used with parenthesis (with or without argument)
        if procedure is None:
            return decorated

        # If @server.register_procedure is used without a parenthesis
        return decorated(procedure)

    @property
    def procedures(self) -> dict[str, ProcedureWrapper]:
        return self._registry


class RpcNamespace(RegistryMixin): ...


class RpcServer(RegistryMixin):
    def __init__(
        self,
        register_system_procedures: bool = True,
        supported_protocol: Protocol = Protocol.ALL,
        auth: AuthPredicateType = NOT_SET,
        error_handler: RpcErrorHandler | None = None,
    ) -> None:
        super().__init__(auth)
        self.request_handlers_classes = list(
            filter(
                lambda cls: check_flags_compatibility(cls.protocol, supported_protocol),
                (import_string(klass) for klass in settings.MODERNRPC_HANDLERS),
            )
        )

        if register_system_procedures:
            system_namespace = import_string(SYSTEM_NAMESPACE_DOTTED_PATH)
            self.register_namespace(system_namespace, "system")

        self.error_handler = error_handler

    def register_namespace(self, namespace: RpcNamespace, name: str | None = None) -> None:
        prefix = f"{name}." if name else ""
        logger.debug(f"About to register {len(namespace.procedures)} procedures into namespace '{prefix}'")

        for procedure_name, wrapper in namespace.procedures.items():
            self.register_procedure(
                wrapper.func_or_coro,
                name=f"{prefix}{procedure_name}",
                protocol=wrapper.protocol,
                auth=wrapper.auth,
                context_target=wrapper.context_target,
            )

    def get_procedure_wrapper(self, name: str, protocol: Protocol) -> ProcedureWrapper:
        """Return the procedure wrapper with given name compatible with given protocol, or raise RPCMethodNotFound"""
        try:
            wrapper = self.procedures[name]
        except KeyError:
            raise RPCMethodNotFound(name) from None

        if check_flags_compatibility(wrapper.protocol, protocol):
            return wrapper

        raise RPCMethodNotFound(name) from None

    def get_request_handler(self, request: HttpRequest) -> RpcHandler | None:
        klass = first_true(self.request_handlers_classes, pred=lambda cls: cls.can_handle(request))
        try:
            return klass()
        except TypeError:
            return None

    def on_error(self, exception: BaseException, context: RpcRequestContext) -> RPCException:
        """
        If an error handler is defined, call it to run arbitrary code and return its not-None result.
        Else, check the given exception type and return it if it is a subclass of RPCException.
        Convert any other exception into RPCInternalError.
        """
        if self.error_handler:
            try:
                self.error_handler(exception, context)
            except Exception as exc:
                exception = exc

        if isinstance(exception, RPCException):
            return exception
        return RPCInternalError(message=str(exception))

    @property
    def view(self) -> Callable:
        """
        Returns a synchronous view function that can be used in Django URL patterns.
        The view is decorated with csrf_exempt and require_POST.

        :return: A callable view function
        """
        view_func = functools.partial(handle_rpc_request, server=self)
        return csrf_exempt(require_POST(view_func))

    @property
    def async_view(self) -> Callable:
        """
        Returns an asynchronous view function that can be used in Django URL patterns.
        The view is decorated with csrf_exempt and require_POST.

        :return: An awaitable async view function
        """
        view_func = functools.partial(handle_rpc_request_async, server=self)
        return async_csrf_exempt(async_require_post(view_func))



from __future__ import annotations

import logging
from http import HTTPStatus
from typing import TYPE_CHECKING

from django.http.response import HttpResponse

from modernrpc.config import settings
from modernrpc.core import RpcRequestContext

if TYPE_CHECKING:
    from django.http import HttpRequest

    from modernrpc.server import RpcServer


logger = logging.getLogger(__name__)


def handle_rpc_request(
    request: HttpRequest, server: RpcServer, default_encoding: str = settings.MODERNRPC_DEFAULT_ENCODING
) -> HttpResponse:
    """
    Synchronous view function to handle RPC requests.

    :param request: The HTTP request object
    :param server: The RPC server instance
    :param default_encoding: The default encoding to use for request body
    :return: An HTTP response object
    """
    if not request.content_type:
        return HttpResponse(
            "Unable to handle your request, the Content-Type header is mandatory to allow server "
            "to determine which handler can interpret your request.",
            status=HTTPStatus.BAD_REQUEST,
            content_type="text/plain",
        )

    handler = server.get_request_handler(request)
    if not handler:
        return HttpResponse(
            f"Unable to handle your request, unsupported Content-Type {request.content_type}.",
            status=HTTPStatus.BAD_REQUEST,
            content_type=request.content_type,
        )

    context = RpcRequestContext(request, server, handler, handler.protocol)
    request_body = request.body.decode(request.encoding or default_encoding)
    result_data = handler.process_request(request_body, context)

    if isinstance(result_data, tuple) and len(result_data) == 2:
        status, result_data = result_data
    else:
        status = HTTPStatus.OK

    return HttpResponse(result_data, status=status, content_type=handler.response_content_type)


async def handle_rpc_request_async(
    request: HttpRequest, server: RpcServer, default_encoding: str = settings.MODERNRPC_DEFAULT_ENCODING
) -> HttpResponse:
    """
    Asynchronous view function to handle RPC requests.

    :param request: The HTTP request object
    :param server: The RPC server instance
    :param default_encoding: The default encoding to use for request body
    :return: An HTTP response object
    """
    if not request.content_type:
        return HttpResponse(
            "Unable to handle your request, the Content-Type header is mandatory to allow server "
            "to determine which handler can interpret your request.",
            status=HTTPStatus.BAD_REQUEST,
            content_type="text/plain",
        )

    handler = server.get_request_handler(request)

    if not handler:
        return HttpResponse(
            f"Unable to handle your request, unsupported Content-Type {request.content_type}.",
            status=HTTPStatus.BAD_REQUEST,
            content_type=request.content_type,
        )

    context = RpcRequestContext(request, server, handler, handler.protocol)
    request_body = request.body.decode(request.encoding or default_encoding)
    result_data = await handler.aprocess_request(request_body, context)

    if isinstance(result_data, tuple) and len(result_data) == 2:
        status, result_data = result_data
    else:
        status = HTTPStatus.OK

    return HttpResponse(result_data, status=status, content_type=handler.response_content_type)



from modernrpc.constants import Protocol
from modernrpc.core import RpcRequestContext
from modernrpc.server import RpcNamespace, RpcServer

__all__ = [
    "Protocol",
    "RpcNamespace",
    "RpcRequestContext",
    "RpcServer",
]



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

import warnings
from functools import cached_property
from typing import TYPE_CHECKING, Any, Iterable

import orjson

from modernrpc.exceptions import RPCMarshallingError, RPCParseError
from modernrpc.jsonrpc.backends.marshalling import Marshaller, Unmarshaller

if TYPE_CHECKING:
    from modernrpc.jsonrpc.handler import JsonRpcRequest, JsonRpcResult


class OrjsonBackend:
    """json-rpc serializer and deserializer based on the third-party orjson library"""

    def __init__(self, load_kwargs: dict[str, Any] | None = None, dump_kwargs: dict[str, Any] | None = None):
        if load_kwargs:
            warnings.warn(
                message="OrJSON backend does not support 'load_kwargs' argument, it will be ignored.",
                category=UserWarning,
                stacklevel=3,
            )

        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def unmarshaller(self):
        return Unmarshaller()

    @cached_property
    def marshaller(self):
        return Marshaller()

    def loads(self, data: str) -> JsonRpcRequest | list[JsonRpcRequest]:
        try:
            structured_data: list[dict] | dict[str, Any] = orjson.loads(data)
        except orjson.JSONDecodeError as exc:
            raise RPCParseError(exc.msg, data=exc) from exc

        return self.unmarshaller.dict_to_request(structured_data)

    def dumps(self, result: JsonRpcResult | Iterable[JsonRpcResult]) -> str:
        structured_data = self.marshaller.result_to_dict(result)
        try:
            return orjson.dumps(structured_data, **self.dump_kwargs).decode("utf-8")
        except (orjson.JSONEncodeError, UnicodeDecodeError, TypeError) as exc:
            raise RPCMarshallingError(structured_data, exc) from exc



"""Handle all introspection-related features on rpc_methods:
- Help text
- Signature
"""

from __future__ import annotations

import importlib
import inspect
import re
from typing import Callable

from django.utils.functional import cached_property

from modernrpc.config import settings

# Define regular expressions used to parse docstring
PARAM_REXP = re.compile(r"^[:@]param (\w+): ?(.*(?:\n[^:@].+)?)$", flags=re.MULTILINE)
RETURN_REXP = re.compile(r"^[:@]return: ?(.*(?:\n[^:@].+)?)$", flags=re.MULTILINE)

PARAM_TYPE_REXP = re.compile(r"^[:@]type (\w+): ?(.*)$", flags=re.MULTILINE)
RETURN_TYPE_REXP = re.compile(r"^[:@]rtype ?: ?(.*)$", flags=re.MULTILINE)

MULTI_SPACES = re.compile(r"\s+")


class Introspector:
    """Helper to extract the signature and type hint for the given function"""

    def __init__(self, function: Callable):
        self.func = function

    @cached_property
    def signature(self) -> inspect.Signature:
        return inspect.signature(self.func)

    @cached_property
    def accept_kwargs(self) -> bool:
        """Determine if function signature contains **kwargs special argument"""
        if self.signature.parameters:
            last_param = next(reversed(list(self.signature.parameters.values())))
            return last_param.kind == inspect.Parameter.VAR_KEYWORD
        return False

    @cached_property
    def args(self) -> list[str]:
        """List all function arguments"""
        args = list(self.signature.parameters.keys())
        if self.accept_kwargs:
            return args[:-1]
        return args

    @cached_property
    def return_type(self) -> str:
        """Return the return type as defined from the signature typehint"""
        return (
            ""
            if self.signature.return_annotation == self.signature.empty
            else getattr(
                self.signature.return_annotation,
                "__name__",
                str(self.signature.return_annotation),
            )
        )

    @cached_property
    def args_types(self) -> dict[str, str]:
        if not self.signature.parameters:
            return {}
        return {
            name: getattr(param.annotation, "__name__", str(param.annotation))
            for name, param in self.signature.parameters.items()
            if param.annotation != param.empty
        }


class DocstringParser:
    """Parse docstring to extract documentation (type and text) for parameters and return. It also converts
    the docstring text part to an HTML representation using Markdown or Restructured parser (depending on settings)
    """

    def __init__(self, function: Callable):
        self.func = function

    @cached_property
    def raw_docstring(self) -> str:
        """Return full content of function docstring, as extracted by inspect.getdoc()"""
        return inspect.getdoc(self.func) or ""

    @staticmethod
    def cleanup_spaces(subject: str) -> str:
        """Basically, replaces any number of consecutive space chars (space, newline, tab, etc.) to a single space"""
        return MULTI_SPACES.sub(" ", subject)

    @cached_property
    def text_documentation(self) -> str:
        """Return the text part of the docstring block, excluding legacy typehints, parameters and return docs"""
        content = self.raw_docstring
        for pattern in [PARAM_REXP, PARAM_TYPE_REXP, RETURN_REXP, RETURN_TYPE_REXP]:
            content = pattern.sub("", content)
        return content.strip()

    @cached_property
    def html_documentation(self) -> str:
        """Convert the text part of the docstring to an HTML representation, using the parser set in settings"""
        if not self.text_documentation:
            return ""

        if settings.MODERNRPC_DOC_FORMAT.lower() in ("rst", "restructured", "restructuredtext"):
            docutils = importlib.import_module("docutils.core")
            return docutils.publish_parts(self.text_documentation, writer_name="html")["body"]

        if settings.MODERNRPC_DOC_FORMAT.lower() in ("md", "markdown"):
            markdown_module = importlib.import_module("markdown")
            return markdown_module.markdown(self.text_documentation)

        html_content = self.text_documentation.replace("\n\n", "</p><p>").replace("\n", " ")
        return f"<p>{html_content}</p>"

    @cached_property
    def args_doc(self) -> dict[str, str]:
        """Return a dict with, for each argument, its name as key and documentation text as value. The dict
        will contain only documented arguments.
        Basically this method parses and extracts reST documented arguments:

            :param <argname>: Documentation for <argname>

        Alternatively, it also supports the @param format:

            @param <argname>: Documentation for <argname>
        """
        return {param: self.cleanup_spaces(text) for param, text in PARAM_REXP.findall(self.raw_docstring)}

    @cached_property
    def args_types(self) -> dict[str, str]:
        """Return a dict with, for each argument, its name as key and documented type as value. The dict
        will contain only documented arguments.
        Basically this method parses and extracts reST doctype documentation:

            :type <argname>: int or str

        Alternatively, it also supports the @type format:

            @type <argname>: int or str
        """
        return dict(PARAM_TYPE_REXP.findall(self.raw_docstring))

    @cached_property
    def return_doc(self) -> str:
        """Return the documentation for method return value, as found in docstring:

            :return: Documentation on return value

        Alternatively, it also supports @return format:

            @return: Documentation on return value

        Return an empty string if return documentation is not found
        """
        res = RETURN_REXP.search(self.raw_docstring)
        return self.cleanup_spaces(res.group(1)) if res else ""

    @cached_property
    def return_type(self) -> str:
        """Return the documentation for the method return type, as found in docstring:

            :rtype: list

        Alternatively, it also supports the @rtype format:

            @rtype: list

        Return an empty string if return type information is not found
        """
        res = RETURN_TYPE_REXP.search(self.raw_docstring)
        return res.group(1) if res else ""



# PEP 585: use of list[Any] instead of List[Any] is available since Python 3.9, enable it for older versions
# PEP 604: use of typeA | typeB is available since Python 3.10, enable it for older versions
from __future__ import annotations

import base64
import xml.parsers.expat
from collections import OrderedDict
from datetime import datetime
from functools import cached_property
from typing import TYPE_CHECKING, Any, Callable, Literal, Sequence

import defusedxml.ElementTree
import xmltodict

from modernrpc.exceptions import RPCInsecureRequest, RPCInvalidRequest, RPCMarshallingError, RPCParseError
from modernrpc.helpers import ensure_sequence, first
from modernrpc.types import RpcErrorResult
from modernrpc.xmlrpc.backends.constants import MAXINT, MININT
from modernrpc.xmlrpc.handler import XmlRpcRequest

try:
    # types.NoneType is available only with Python 3.10+
    from types import NoneType
except ImportError:
    NoneType = type(None)  # type: ignore[misc]


if TYPE_CHECKING:
    from modernrpc.xmlrpc.handler import XmlRpcResult


LoadFuncType = Callable[[Any], Any]
DumpFuncType = Callable[[Any], dict]


class Unmarshaller:
    def __init__(self) -> None:
        self.load_funcs: dict[str, LoadFuncType] = {
            "value": self.load_value,
            "nil": self.load_nil,
            "boolean": self.load_bool,
            "int": self.load_int,
            "i4": self.load_int,
            "double": self.load_float,
            "string": self.load_str,
            "dateTime.iso8601": self.load_datetime,
            "base64": self.load_base64,
            "array": self.load_array,
            "struct": self.load_struct,
        }

    def dict_to_request(self, data: OrderedDict[str, Any]) -> XmlRpcRequest:
        try:
            method_call = data["methodCall"]
        except KeyError as exc:
            raise RPCInvalidRequest("missing methodCall tag", data=data) from exc

        try:
            method_name = method_call["methodName"]
        except KeyError as exc:
            raise RPCInvalidRequest("missing methodCall.methodName tag", data=data) from exc

        params = method_call.get("params") or {}
        param_list: Sequence[dict[str, Any]] = params.get("param", [])

        args: list[Any] = []
        if len(param_list) == 0:
            args = []
        elif len(param_list) == 1:
            param_list = ensure_sequence(param_list)

        for param in param_list:
            _type, v = first(param.items())
            args.append(self.dispatch(_type, v))

        return XmlRpcRequest(method_name=method_name, args=args)

    def dispatch(self, _type: str, value: Any) -> Any:
        try:
            load_func = self.load_funcs[_type]
        except KeyError as exc:
            raise RPCInvalidRequest(f"Unsupported type {_type}") from exc

        return load_func(value)

    def load_value(self, data: dict) -> Any:
        _type, v = first(data.items())
        return self.dispatch(_type, v)

    @staticmethod
    def load_nil(_) -> None:
        return None

    @staticmethod
    def load_int(data: str) -> int:
        return int(data)

    @staticmethod
    def load_bool(data: str) -> bool:
        if data not in ("0", "1"):
            raise TypeError("Invalid boolean value: only 0 and 1 are allowed")
        return data == "1"

    @staticmethod
    def load_float(data: str) -> float:
        return float(data)

    @staticmethod
    def load_str(data: str) -> str:
        return str(data)

    @staticmethod
    def load_datetime(data: str) -> datetime:
        return datetime.strptime(data, "%Y%m%dT%H:%M:%S")

    @staticmethod
    def load_base64(data: str) -> bytes:
        return base64.b64decode(data)

    def load_array(self, data: dict[str, dict[str, list[dict[str, Any]]]]):
        entries = []
        for element in ensure_sequence(data["data"]["value"]):
            _type, value = first(element.items())
            entries.append(self.dispatch(_type, value))
        return entries

    def load_struct(self, data: dict):
        members = ensure_sequence(data["member"])
        res = {}
        for member in members:
            value = member["value"]
            if len(value) > 1:
                raise ValueError
            _type, value = first(value.items())
            res[member["name"]] = self.dispatch(_type, value)
        return res


class Marshaller:
    def __init__(self) -> None:
        self.dump_funcs: dict[type, DumpFuncType] = {
            NoneType: self.dump_nil,
            bool: self.dump_bool,
            int: self.dump_int,
            float: self.dump_float,
            str: self.dump_str,
            bytes: self.dump_bytearray,
            bytearray: self.dump_bytearray,
            datetime: self.dump_datetime,
            list: self.dump_list,
            tuple: self.dump_list,
            dict: self.dump_dict,
            OrderedDict: self.dump_dict,
        }

    def result_to_dict(self, result: XmlRpcResult) -> dict[str, Any]:
        if isinstance(result, RpcErrorResult):
            return {
                "methodResponse": {
                    "fault": {
                        "value": self.dispatch(
                            {
                                "faultCode": result.code,
                                "faultString": result.message,
                            }
                        ),
                    }
                }
            }

        return {
            "methodResponse": {
                "params": [
                    {"param": {"value": self.dispatch(result.data)}},
                ]
            }
        }

    def dispatch(self, value: Any) -> dict[str, Any]:
        try:
            dump_func = self.dump_funcs[type(value)]
        except KeyError as e:
            raise TypeError(f"Unsupported type: {type(value)}") from e
        return dump_func(value)

    @staticmethod
    def dump_nil(_):
        # FIXME: allow_none config ?
        return {"nil": None}

    @staticmethod
    def dump_bool(value: bool) -> dict[str, Literal[1, 0]]:
        return {"boolean": 1 if value else 0}

    @staticmethod
    def dump_int(value: int) -> dict[str, int]:
        if value > MAXINT or value < MININT:
            raise OverflowError("int value exceeds XML-RPC limits")
        return {"int": value}

    @staticmethod
    def dump_float(value: float) -> dict[str, float]:
        return {"double": value}

    @staticmethod
    def dump_str(value: str) -> dict[str, str]:
        return {"string": value}

    @staticmethod
    def dump_datetime(value: datetime) -> dict[str, str]:
        return {"dateTime.iso8601": value.strftime("%04Y%02m%02dT%H:%M:%S")}

    @staticmethod
    def dump_bytearray(value: bytes | bytearray) -> dict[str, str]:
        return {"base64": base64.b64encode(value).decode()}

    def dump_dict(self, value: dict) -> dict[str, dict[str, list[dict[str, Any]]]]:
        return {
            "struct": {
                "member": [
                    {
                        "name": key,
                        "value": self.dispatch(val),
                    }
                    for key, val in value.items()
                ],
            },
        }

    def dump_list(self, value: list | tuple) -> dict[str, dict[str, dict[str, list[Any]]]]:
        return {
            "array": {
                "data": {
                    "value": [self.dispatch(val) for val in value],
                }
            },
        }


class XmlToDictBackend:
    """xml-rpc serializer and deserializer based on the third-party xmltodict library"""

    def __init__(self, load_kwargs: dict[str, Any] | None = None, dump_kwargs: dict[str, Any] | None = None):
        self.load_kwargs = load_kwargs or {}
        self.dump_kwargs = dump_kwargs or {}

    @cached_property
    def marshaller(self):
        return Marshaller()

    @cached_property
    def unmarshaller(self):
        return Unmarshaller()

    def loads(self, data: str) -> XmlRpcRequest:
        try:
            # First, parse the XML using defusedxml.ElementTree for security
            # This will raise appropriate exceptions for XML vulnerabilities
            root = defusedxml.ElementTree.fromstring(data)

            # Convert the parsed XML to a string and then parse it with xmltodict
            # This is safe because defusedxml has already validated the XML
            data = defusedxml.ElementTree.tostring(root, encoding="utf-8").decode("utf-8")
        except defusedxml.ElementTree.ParseError as exc:
            raise RPCParseError(str(exc)) from exc
        except defusedxml.DefusedXmlException as exc:
            raise RPCInsecureRequest(str(exc)) from exc

        try:
            structured_data: OrderedDict[str, Any] = xmltodict.parse(data, **self.load_kwargs)
        except xml.parsers.expat.ExpatError as exc:
            raise RPCParseError(str(exc)) from exc

        try:
            return self.unmarshaller.dict_to_request(structured_data)
        except TypeError as exc:
            raise RPCInvalidRequest(str(exc)) from exc

    def dumps(self, result: XmlRpcResult) -> str:
        try:
            structured_data = self.marshaller.result_to_dict(result)
        except Exception as e:
            raise RPCMarshallingError(result.data, e) from e

        return xmltodict.unparse(structured_data, **self.dump_kwargs)
